import subprocess
import time
from uuid import uuid4

import pytest
from thrift.transport import TSocket, TTransport
from thrift.transport.TSocket import TTransportException
from pymapd import connect
import datetime
import random
import string
import numpy as np
import pandas as pd


def _check_open():
    """
    Test to see if OmniSci running on localhost and socket open
    """
    socket = TSocket.TSocket("localhost", 6274)
    transport = TTransport.TBufferedTransport(socket)

    try:
        transport.open()
        return True
    except TTransportException:
        return False


@pytest.fixture(scope='session')
def mapd_server():
    """Ensure a mapd server is running, optionally starting one if none"""
    if _check_open():
        # already running before pytest started
        pass
    else:
        # not yet running...
        subprocess.check_output(
            [
                'docker',
                'run',
                '-d',
                '--ipc=host',
                '-v',
                '/dev:/dev',
                '-p',
                '6274:6274',
                '-p',
                '9092:9092',
                '--name=mapd',
                'omnisci/core-os-cpu:latest',
            ]
        )
        # yield and stop afterwards?
        assert _check_open()
        # Takes some time to start up. Unfortunately even trying to connect
        # will cause it to hang.
        time.sleep(5)


@pytest.fixture(scope='session')
def con(mapd_server):
    """
    Fixture to provide Connection for tests run against live OmniSci instance
    """
    return connect(
        user="admin",
        password='HyperInteractive',
        host='localhost',
        port=6274,
        protocol='binary',
        dbname='omnisci',
    )


@pytest.fixture
def mock_client(mocker):
    """A magicmock for pymapd.connection.Client"""
    return mocker.patch("pymapd.connection.Client")


def no_gpu():
    """Check for the required GPU dependencies"""
    try:
        from numba import cuda
        import cudf  # noqa

        try:
            cuda.select_device(0)
        except cuda.cudadrv.error.CudaDriverError:
            return True
    except ImportError:
        return True
    return False


def gen_string():
    """Generate a random string sequence for use in _tests_table_no_nulls"""
    return ''.join(
        [
            random.choice(string.ascii_letters + string.digits)
            for n in range(10)
        ]
    )


def _tests_table_no_nulls(n_samples):
    """
    Generates a dataframe with all OmniSci types in it for use in integration
    testing
    """

    np.random.seed(12345)

    tinyint_ = np.random.randint(
        low=-127, high=127, size=n_samples, dtype='int8'
    )

    smallint_ = np.random.randint(
        low=-32767, high=32767, size=n_samples, dtype='int16'
    )

    int_ = np.random.randint(
        low=-2147483647, high=2147483647, size=n_samples, dtype='int32'
    )

    bigint_ = np.random.randint(
        low=-9223372036854775807,
        high=9223372036854775807,
        size=n_samples,
        dtype='int64',
    )

    # float and double ranges slightly lower than we support, full width
    # causes an error in np.linspace that's not worth tracking down
    float_ = np.linspace(-3.4e37, 3.4e37, n_samples, dtype='float32')
    double_ = np.linspace(-1.79e307, 1.79e307, n_samples, dtype='float64')

    bool_ = np.random.randint(low=0, high=2, size=n_samples, dtype='bool')

    # effective date range of 1904 to 2035
    # TODO: validate if this is an Arrow limitation, outside this range fails
    date_ = [
        datetime.date(1970, 1, 1) + datetime.timedelta(days=int(x))
        for x in np.random.randint(-24000, 24000, size=n_samples)
    ]

    datetime_ = [
        datetime.datetime(1970, 1, 1)
        + datetime.timedelta(days=int(x), minutes=int(x))
        for x in np.random.randint(-24000, 24000, size=n_samples)
    ]

    time_h = np.random.randint(0, 24, size=n_samples)
    time_m = np.random.randint(0, 60, size=n_samples)
    time_s = np.random.randint(0, 60, size=n_samples)
    time_ = [datetime.time(h, m, s) for h, m, s in zip(time_h, time_m, time_s)]

    # generate random text strings
    text_ = [gen_string() for x in range(n_samples)]

    # read geo data from files
    point_ = pd.read_csv("tests/data/points_10000.zip", header=None).values
    point_ = np.squeeze(point_)

    line_ = pd.read_csv("tests/data/lines_10000.zip", header=None).values
    line_ = np.squeeze(line_)

    mpoly_ = pd.read_csv("tests/data/mpoly_10000.zip", header=None).values
    mpoly_ = np.squeeze(mpoly_)

    poly_ = pd.read_csv("tests/data/polys_10000.zip", header=None).values
    poly_ = np.squeeze(poly_)

    d = {
        'tinyint_': tinyint_,
        'smallint_': smallint_,
        'int_': int_,
        'bigint_': bigint_,
        'float_': float_,
        'double_': double_,
        'bool_': bool_,
        'date_': date_,
        'datetime_': datetime_,
        'time_': time_,
        'text_': text_,
        'point_': point_,
        'line_': line_,
        'mpoly_': mpoly_,
        'poly_': poly_,
    }

    return pd.DataFrame(d)


@pytest.fixture
def tmp_table(con) -> str:
    table_name = 'table_{}'.format(uuid4().hex)
    con.execute("drop table if exists {};".format(table_name))

    try:
        yield table_name
    finally:
        con.execute("drop table if exists {};".format(table_name))



"""
Utility methods for parsing data returned from MapD
"""
import datetime
from collections import namedtuple
from sqlalchemy import text
import omnisci.common.ttypes as T
import ctypes
from types import MethodType
from ._mutators import set_tdf, get_tdf
from ._utils import seconds_to_time, datetime_in_precisions
import numpy as np
from .ipc import load_buffer, shmdt


Description = namedtuple(
    "Description",
    [
        "name",
        "type_code",
        "display_size",
        "internal_size",
        "precision",
        "scale",
        "null_ok",
    ],
)
ColumnDetails = namedtuple(
    "ColumnDetails",
    [
        "name",
        "type",
        "nullable",
        "precision",
        "scale",
        "comp_param",
        "encoding",
        "is_array",
    ],
)

_typeattr = {
    'SMALLINT': 'int',
    'INT': 'int',
    'BIGINT': 'int',
    'TIME': 'int',
    'TIMESTAMP': 'int',
    'DATE': 'int',
    'BOOL': 'int',
    'FLOAT': 'real',
    'DECIMAL': 'real',
    'DOUBLE': 'real',
    'STR': 'str',
    'POINT': 'str',
    'LINESTRING': 'str',
    'POLYGON': 'str',
    'MULTIPOLYGON': 'str',
    'TINYINT': 'int',
    'GEOMETRY': 'str',
    'GEOGRAPHY': 'str',
}
_thrift_types_to_values = T.TDatumType._NAMES_TO_VALUES
_thrift_values_to_types = T.TDatumType._VALUES_TO_NAMES
_thrift_encodings_to_values = T.TEncodingType._NAMES_TO_VALUES
_thrift_values_to_encodings = T.TEncodingType._VALUES_TO_NAMES


def _format_result_timestamp(desc, arr):

    return [
        None
        if v is None
        else datetime_in_precisions(v, desc.col_type.precision)
        for v in arr
    ]


def _format_result_date(arr):

    base = datetime.datetime(1970, 1, 1)
    return [
        None if v is None else (base + datetime.timedelta(seconds=v)).date()
        for v in arr
    ]


def _format_result_time(arr):

    return [None if v is None else seconds_to_time(v) for v in arr]


def _extract_col_vals(desc, val):

    typename = T.TDatumType._VALUES_TO_NAMES[desc.col_type.type]
    nulls = val.nulls

    # arr_col has multiple levels to parse, not accounted for in original code
    # https://github.com/omnisci/pymapd/issues/68
    if hasattr(val.data, 'arr_col') and val.data.arr_col:
        vals = [
            None if null else getattr(v.data, _typeattr[typename] + '_col')
            for null, v in zip(nulls, val.data.arr_col)
        ]

        if typename == 'TIMESTAMP':
            vals = [_format_result_timestamp(desc, v) for v in vals]
        elif typename == 'DATE':
            vals = [_format_result_date(v) for v in vals]
        elif typename == 'TIME':
            vals = [_format_result_time(v) for v in vals]

    # else clause original code path
    else:
        vals = getattr(val.data, _typeattr[typename] + '_col')
        vals = [None if null else v for null, v in zip(nulls, vals)]

        if typename == 'TIMESTAMP':
            vals = _format_result_timestamp(desc, vals)
        elif typename == 'DATE':
            vals = _format_result_date(vals)
        elif typename == 'TIME':
            vals = _format_result_time(vals)

    return vals


def _extract_description(row_desc):
    """
    Return a tuple of (name, type_code, display_size, internal_size,
                       precision, scale, null_ok)

    https://www.python.org/dev/peps/pep-0249/#description
    """
    return [
        Description(
            col.col_name,
            col.col_type.type,
            None,
            None,
            None,
            None,
            col.col_type.nullable,
        )
        for col in row_desc
    ]


def _extract_column_details(row_desc):
    # For Connection.get_table_details
    return [
        ColumnDetails(
            x.col_name,
            _thrift_values_to_types[x.col_type.type],
            x.col_type.nullable,
            x.col_type.precision,
            x.col_type.scale,
            x.col_type.comp_param,
            _thrift_values_to_encodings[x.col_type.encoding],
            x.col_type.is_array,
        )
        for x in row_desc
    ]


def _parse_tdf_gpu(tdf):
    """
    Parse the results of a select ipc_gpu into a GpuDataFrame

    Parameters
    ----------
    tdf : TDataFrame

    Returns
    -------
    gdf : GpuDataFrame
    """

    import pyarrow as pa
    from cudf.comm.gpuarrow import GpuArrowReader
    from cudf.core.dataframe import DataFrame
    from pyarrow._cuda import Context, IpcMemHandle
    from numba import cuda

    ipc_handle = IpcMemHandle.from_buffer(pa.py_buffer(tdf.df_handle))
    ctx = Context()
    ipc_buf = ctx.open_ipc_buffer(ipc_handle)
    ipc_buf.context.synchronize()

    schema_buffer, shm_ptr = load_buffer(tdf.sm_handle, tdf.sm_size)

    buffer = pa.BufferReader(schema_buffer)
    schema = pa.ipc.read_schema(buffer)

    # Dictionary Memo functionality used to
    # deserialize on the C++ side is not
    # exposed on the pyarrow side, so we need to
    # handle this on our own.
    dict_memo = {}

    try:
        dict_batch_reader = pa.RecordBatchStreamReader(buffer)
        updated_fields = []

        for f in schema:
            if pa.types.is_dictionary(f.type):
                msg = dict_batch_reader.read_next_batch()
                dict_memo[f.name] = msg.column(0)
                updated_fields.append(pa.field(f.name, f.type.index_type))
            else:
                updated_fields.append(pa.field(f.name, f.type))

        schema = pa.schema(updated_fields)
    except pa.ArrowInvalid:
        # This message does not have any dictionary encoded
        # columns
        pass

    dtype = np.dtype(np.byte)
    darr = cuda.devicearray.DeviceNDArray(
        shape=ipc_buf.size,
        strides=dtype.itemsize,
        dtype=dtype,
        gpu_data=ipc_buf.to_numba(),
    )

    reader = GpuArrowReader(schema, darr)
    df = DataFrame()
    df.set_tdf = MethodType(set_tdf, df)
    df.get_tdf = MethodType(get_tdf, df)

    for k, v in reader.to_dict().items():
        if k in dict_memo:
            df[k] = pa.DictionaryArray.from_arrays(v.to_arrow(), dict_memo[k])
        else:
            df[k] = v

    df.set_tdf(tdf)

    # free shared memory from Python
    # https://github.com/omnisci/pymapd/issues/46
    # https://github.com/omnisci/pymapd/issues/31
    free_sm = shmdt(ctypes.cast(shm_ptr, ctypes.c_void_p))  # noqa

    return df


def _bind_parameters(operation, parameters):
    return (
        text(operation)
        .bindparams(**parameters)
        .compile(compile_kwargs={"literal_binds": True})
    )



"""
Tests that rely on a server running
"""
import base64
import json
import datetime
from unittest import mock

import pytest
from pymapd import connect, ProgrammingError, DatabaseError
from pymapd.cursor import Cursor
from pymapd._parsers import Description, ColumnDetails
from omnisci.thrift.ttypes import TOmniSciException
from omnisci.common.ttypes import TDatumType
from .data import dashboard_metadata

import geopandas as gpd
import pandas as pd
import numpy as np
import pyarrow as pa
from pandas.api.types import is_object_dtype, is_categorical_dtype
import pandas.testing as tm
import shapely
from shapely.geometry import Point, LineString, Polygon, MultiPolygon
import textwrap
from .conftest import no_gpu

# XXX: Make it hashable to silence warnings; see if this can be done upstream
# This isn't a huge deal, but our testing context mangers for asserting
# exceptions need hashability
TOmniSciException.__hash__ = lambda x: id(x)


def _cursor2df(cursor):
    col_types = {c.name: c.type_code for c in cursor.description}
    has_geodata = {
        k: v
        in [
            TDatumType.POINT,
            TDatumType.LINESTRING,
            TDatumType.POLYGON,
            TDatumType.MULTIPOLYGON,
        ]
        for k, v in col_types.items()
    }
    col_names = list(col_types.keys())

    df_class = gpd.GeoDataFrame if any(has_geodata.values()) else pd.DataFrame
    df = df_class(cursor.fetchall(), columns=col_names)

    for c, _has_geodata in has_geodata.items():
        if _has_geodata:
            df.loc[:, c] = df.loc[:, c].apply(shapely.wkt.loads)
    return df


@pytest.mark.usefixtures("mapd_server")
class TestIntegration:
    def test_connect_binary(self):
        con = connect(
            user="admin",
            password='HyperInteractive',
            host='localhost',
            port=6274,
            protocol='binary',
            dbname='omnisci',
        )
        assert con is not None

    def test_connect_http(self):
        con = connect(
            user="admin",
            password='HyperInteractive',
            host='localhost',
            port=6278,
            protocol='http',
            dbname='omnisci',
        )
        assert con is not None

    def test_connect_uri(self):
        uri = (
            'omnisci://admin:HyperInteractive@localhost:6274/omnisci?'
            'protocol=binary'
        )
        con = connect(uri=uri)
        assert con._user == 'admin'
        assert con._password == 'HyperInteractive'
        assert con._host == 'localhost'
        assert con._port == 6274
        assert con._dbname == 'omnisci'
        assert con._protocol == 'binary'

    def test_connect_uri_and_others_raises(self):
        uri = (
            'omnisci://admin:HyperInteractive@localhost:6274/omnisci?'
            'protocol=binary'
        )
        with pytest.raises(TypeError):
            connect(username='omnisci', uri=uri)

    def test_invalid_sql(self, con):
        with pytest.raises(ProgrammingError) as r:
            con.cursor().execute("this is invalid;")
        r.match("Exception: Parse failed:")

    def test_nonexistant_table(self, con):
        with pytest.raises(DatabaseError) as r:
            con.cursor().execute("select it from fake_table;")
        r.match("Table 'FAKE_TABLE' does not exist|Object 'fake_table' not")

    def test_connection_execute(self, con):
        result = con.execute("drop table if exists FOO;")
        result = con.execute("create table FOO (a int);")
        assert isinstance(result, Cursor)
        con.execute("drop table if exists FOO;")

    def test_select_sets_description(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute("select * from stocks")
        expected = [
            Description('date_', 6, None, None, None, None, True),
            Description('trans', 6, None, None, None, None, True),
            Description('symbol', 6, None, None, None, None, True),
            Description('qty', 1, None, None, None, None, True),
            Description('price', 3, None, None, None, None, True),
            Description('vol', 3, None, None, None, None, True),
        ]
        assert c.description == expected
        c.execute('drop table if exists stocks;')

    def test_select_parametrized(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute(
            'select symbol, qty from stocks where symbol = :symbol',
            {'symbol': 'GOOG'},
        )
        result = list(c)
        expected = [
            ('GOOG', 100),
        ]  # noqa
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_executemany_parametrized(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        parameters = [{'symbol': 'GOOG'}, {'symbol': "RHAT"}]
        expected = [[('GOOG', 100)], [('RHAT', 100)]]
        query = 'select symbol, qty from stocks where symbol = :symbol'
        c = con.cursor()
        result = c.executemany(query, parameters)
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_executemany_parametrized_insert(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c = con.cursor()
        c.execute("drop table if exists stocks2;")
        # Create table
        c.execute('CREATE TABLE stocks2 (symbol text, qty int);')
        params = [{"symbol": "GOOG", "qty": 10}, {"symbol": "AAPL", "qty": 20}]
        query = "INSERT INTO stocks2 VALUES (:symbol, :qty);"
        result = c.executemany(query, params)
        assert result == [[], []]  # TODO: not sure if this is standard
        c.execute("drop table stocks2;")
        c.execute('drop table if exists stocks;')

    @pytest.mark.parametrize(
        'query, parameters',
        [
            ('select qty, price from stocks', None),
            ('select qty, price from stocks where qty=:qty', {'qty': 100}),
        ],
    )
    def test_select_ipc_parametrized(self, con, query, parameters):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc(query, parameters=parameters)
        expected = pd.DataFrame(
            {
                "qty": np.array([100, 100], dtype=np.int32),
                "price": np.array(
                    [35.13999938964844, 12.140000343322754], dtype=np.float32
                ),
            }
        )[['qty', 'price']]
        tm.assert_frame_equal(result, expected)
        c.execute('drop table if exists stocks;')

    def test_select_ipc_first_n(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc("select * from stocks", first_n=1)
        assert len(result) == 1
        c.execute('drop table if exists stocks;')

    @pytest.mark.parametrize(
        'query, parameters',
        [
            ('select qty, price from stocks', None),
            ('select qty, price from stocks where qty=:qty', {'qty': 100}),
        ],
    )
    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_select_ipc_gpu(self, con, query, parameters):

        from cudf.core.dataframe import DataFrame

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc_gpu("select qty, price from stocks")
        assert isinstance(result, DataFrame)

        dtypes = dict(qty=np.int32, price=np.float32)
        expected = pd.DataFrame(
            [[100, 35.14], [100, 12.14]], columns=['qty', 'price']
        ).astype(dtypes)

        result = result.to_pandas()[['qty', 'price']]  # column order
        pd.testing.assert_frame_equal(result, expected)
        c.execute('drop table if exists stocks;')

    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_select_text_ipc_gpu(self, con):

        from cudf.core.dataframe import DataFrame

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)

        symbols = set(['GOOG', 'RHAT', 'IBM', 'NVDA'])
        for i, sym in enumerate(symbols):
            stmt = "INSERT INTO stocks VALUES ('2006-01-05_{}','BUY','{}',{},35.{},{}.1);".format(  # noqa
                i, sym, i, i, i
            )  # noqa
            # insert twice so we can test
            # that duplicated text values
            # are deserialized properly
            c.execute(stmt)
            c.execute(stmt)

        result = con.select_ipc_gpu(
            "select trans, symbol, qty, price from stocks"
        )  # noqa
        assert isinstance(result, DataFrame)

        assert len(result) == 8
        assert set(result['trans'].to_arrow()) == set(["BUY"])
        assert set(result['symbol'].to_arrow()) == symbols
        c.execute('drop table if exists stocks;')

    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_select_gpu_first_n(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.select_ipc_gpu("select * from stocks", first_n=1)
        assert len(result) == 1
        c.execute('drop table if exists stocks;')

    def test_fetchone(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute("select symbol, qty from stocks")
        result = c.fetchone()
        expected = ('RHAT', 100)
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_fetchmany(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        c.execute("select symbol, qty from stocks")
        result = c.fetchmany()
        expected = [('RHAT', 100)]
        assert result == expected

        c.execute("select symbol, qty from stocks")
        result = c.fetchmany(size=10)
        expected = [('RHAT', 100), ('GOOG', 100)]
        assert result == expected
        c.execute('drop table if exists stocks;')

    def test_select_dates(self, con):

        c = con.cursor()
        c.execute('drop table if exists dates;')
        c.execute(
            'create table dates (date_ DATE, datetime_ TIMESTAMP, '
            'time_ TIME);'
        )
        i1 = (
            "INSERT INTO dates VALUES ('2006-01-05','2006-01-01T12:00:00',"
            "'12:00:00');"
        )
        i2 = (
            "INSERT INTO dates VALUES ('1901-12-14','1901-12-13T20:45:53',"
            "'23:59:00');"
        )
        c.execute(i1)
        c.execute(i2)

        result = list(c.execute("select * from dates"))
        expected = [
            (
                datetime.date(2006, 1, 5),
                datetime.datetime(2006, 1, 1, 12),
                datetime.time(12),
            ),
            (
                datetime.date(1901, 12, 14),
                datetime.datetime(1901, 12, 13, 20, 45, 53),
                datetime.time(23, 59),
            ),
        ]
        assert result == expected
        c.execute('drop table if exists dates;')

    def test_dashboard_duplication_remap(self, con):
        # This test relies on the test_data_no_nulls_ipc table
        # Setup our testing variables
        old_dashboard_state = dashboard_metadata.old_dashboard_state
        old_dashboard_name = dashboard_metadata.old_dashboard_name
        new_dashboard_name = "new_test"
        meta_data = {"table": "test_data_no_nulls_ipc", "version": "v2"}
        remap = {
            "test_data_no_nulls_ipc": {
                "name": new_dashboard_name,
                "title": new_dashboard_name,
            }
        }
        dashboards = []

        # Create testing dashboard
        try:
            dashboard_id = con._client.create_dashboard(
                session=con._session,
                dashboard_name=old_dashboard_name,
                dashboard_state=(
                    base64.b64encode(
                        json.dumps(old_dashboard_state).encode("utf-8")
                    )
                ),
                image_hash="",
                dashboard_metadata=json.dumps(meta_data),
            )
        except TOmniSciException:
            dashboards = con._client.get_dashboards(con._session)
            for dash in dashboards:
                if dash.dashboard_name == old_dashboard_name:
                    con._client.delete_dashboard(
                        con._session, dash.dashboard_id
                    )
                    break
            dashboard_id = con._client.create_dashboard(
                session=con._session,
                dashboard_name=old_dashboard_name,
                dashboard_state=(
                    base64.b64encode(
                        json.dumps(old_dashboard_state).encode("utf-8")
                    )
                ),
                image_hash="",
                dashboard_metadata=json.dumps(meta_data),
            )

        # Duplicate and remap our dashboard
        try:
            dashboard_id = con.duplicate_dashboard(
                dashboard_id, new_dashboard_name, remap
            )
        except TOmniSciException:
            dashboards = con._client.get_dashboards(con._session)
            for dash in dashboards:
                if dash.dashboard_name == new_dashboard_name:
                    con._client.delete_dashboard(
                        con._session, dash.dashboard_id
                    )
                    break
            dashboard_id = con.duplicate_dashboard(
                dashboard_id, new_dashboard_name, remap
            )

        # Get our new dashboard from the database
        d = con.get_dashboard(dashboard_id=dashboard_id)
        remapped_dashboard = json.loads(
            base64.b64decode(d.dashboard_state).decode()
        )

        # Assert that the table and title changed
        assert remapped_dashboard['dashboard']['title'] == new_dashboard_name

        # Ensure the datasources change
        for key, val in remapped_dashboard['dashboard']['dataSources'].items():
            for col in val['columnMetadata']:
                assert col['table'] == new_dashboard_name


class TestOptionalImports:
    def test_select_gpu(self, con):
        with mock.patch.dict(
            "sys.modules", {"cudf": None, "cudf.core.dataframe": None}
        ):
            with pytest.raises(ImportError) as m:
                con.select_ipc_gpu("select * from foo;")
        assert m.match("The 'cudf' package is required")


class TestExtras:
    def test_sql_validate(self, con):
        from omnisci.common.ttypes import TTypeInfo

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)

        q = "select * from stocks"
        results = con._client.sql_validate(con._session, q)
        col_names = sorted([r.col_name for r in results])
        col_types = [r.col_type for r in results]

        expected_col_names = [
            'date_',
            'price',
            'qty',
            'symbol',
            'trans',
            'vol',
        ]

        expected_types = [
            TTypeInfo(
                type=6,
                encoding=4,
                nullable=True,
                is_array=False,
                precision=0,
                scale=0,
                comp_param=32,
                size=-1,
            ),
            TTypeInfo(
                type=6,
                encoding=4,
                nullable=True,
                is_array=False,
                precision=0,
                scale=0,
                comp_param=32,
                size=-1,
            ),
            TTypeInfo(
                type=6,
                encoding=4,
                nullable=True,
                is_array=False,
                precision=0,
                scale=0,
                comp_param=32,
                size=-1,
            ),
            TTypeInfo(
                type=1,
                encoding=0,
                nullable=True,
                is_array=False,
                precision=0,
                scale=0,
                comp_param=0,
                size=-1,
            ),
            TTypeInfo(
                type=3,
                encoding=0,
                nullable=True,
                is_array=False,
                precision=0,
                scale=0,
                comp_param=0,
                size=-1,
            ),
            TTypeInfo(
                type=3,
                encoding=0,
                nullable=True,
                is_array=False,
                precision=0,
                scale=0,
                comp_param=0,
                size=-1,
            ),
        ]

        assert col_types == expected_types
        assert col_names == expected_col_names

    def test_get_tables(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float);'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1);"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2);"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.get_tables()
        assert isinstance(result, list)
        assert 'stocks' in result
        c.execute('drop table if exists stocks;')

    def test_get_table_details(self, con):

        c = con.cursor()
        c.execute('drop table if exists stocks;')
        create = (
            'create table stocks (date_ text, trans text, symbol text, '
            'qty int, price float, vol float, '
            'exchanges TEXT [] ENCODING DICT(32));'
        )
        c.execute(create)
        i1 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14,1.1,{'NYSE', 'NASDAQ', 'AMEX'});"  # noqa
        i2 = "INSERT INTO stocks VALUES ('2006-01-05','BUY','GOOG',100,12.14,1.2,{'NYSE', 'NASDAQ'});"  # noqa

        c.execute(i1)
        c.execute(i2)

        result = con.get_table_details('stocks')
        expected = [
            ColumnDetails(
                name='date_',
                type='STR',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=32,
                encoding='DICT',
                is_array=False,
            ),
            ColumnDetails(
                name='trans',
                type='STR',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=32,
                encoding='DICT',
                is_array=False,
            ),
            ColumnDetails(
                name='symbol',
                type='STR',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=32,
                encoding='DICT',
                is_array=False,
            ),
            ColumnDetails(
                name='qty',
                type='INT',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=0,
                encoding='NONE',
                is_array=False,
            ),
            ColumnDetails(
                name='price',
                type='FLOAT',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=0,
                encoding='NONE',
                is_array=False,
            ),
            ColumnDetails(
                name='vol',
                type='FLOAT',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=0,
                encoding='NONE',
                is_array=False,
            ),
            ColumnDetails(
                name='exchanges',
                type='STR',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=32,
                encoding='DICT',
                is_array=True,
            ),
        ]
        assert result == expected
        c.execute('drop table if exists stocks;')


class TestLoaders:
    @staticmethod
    def check_empty_insert(result, expected):
        assert len(result) == 3
        assert expected[0][0] == result[0][0]
        assert expected[0][2] == result[0][2]
        assert abs(expected[0][1] - result[0][1]) < 1e-7  # floating point

    def test_load_empty_table(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = [(1, 1.1, 'a'), (2, 2.2, '2'), (3, 3.3, '3')]
        con.load_table("baz", data)
        result = sorted(con.execute("select * from baz"))
        self.check_empty_insert(result, data)
        con.execute("drop table if exists baz;")

    def test_load_empty_table_pandas(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = [(1, 1.1, 'a'), (2, 2.2, '2'), (3, 3.3, '3')]
        df = pd.DataFrame(data, columns=list('abc'))
        con.load_table("baz", df, method='columnar')
        result = sorted(con.execute("select * from baz"))
        self.check_empty_insert(result, data)
        con.execute("drop table if exists baz;")

    def test_load_empty_table_arrow(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = [(1, 1.1, 'a'), (2, 2.2, '2'), (3, 3.3, '3')]

        df = pd.DataFrame(data, columns=list('abc')).astype(
            {'a': 'int32', 'b': 'float32'}
        )

        table = pa.Table.from_pandas(df, preserve_index=False)
        con.load_table("baz", table, method='arrow')
        result = sorted(con.execute("select * from baz"))
        self.check_empty_insert(result, data)
        con.execute("drop table if exists baz;")

    @pytest.mark.parametrize(
        'df, table_fields',
        [
            pytest.param(
                pd.DataFrame(
                    {
                        "a": [1, 2, 3],
                        "b": [1.1, 2.2, 3.3],
                        "c": ['a', '2', '3'],
                    },
                ),
                'a int, b float, c text',
                id='scalar_values',
            ),
            pytest.param(
                pd.DataFrame(
                    {
                        "a": [
                            np.datetime64('2010-01-01 01:01:01.001001001'),
                            np.datetime64('2011-01-01 01:01:01.001001001'),
                            np.datetime64('2012-01-01 01:01:01.001001001'),
                        ],
                    },
                ),
                'a TIMESTAMP(9)',
                id='scalar_datetime_nanoseconds',
            ),
            pytest.param(
                pd.DataFrame(
                    {
                        "a": [
                            datetime.datetime.fromtimestamp(
                                float(1600443582510) / 1e3
                            ),
                            datetime.datetime.fromtimestamp(
                                float(1600443582510) / 1e3
                            ),
                            datetime.datetime.fromtimestamp(
                                float(1600443582510) / 1e3
                            ),
                        ],
                    },
                ),
                'a TIMESTAMP(3)',
                id='scalar_datetime_ms',
            ),
            pytest.param(
                pd.DataFrame(
                    {
                        "a": [
                            datetime.datetime.fromtimestamp(
                                float(1600443582510) / 1e6
                            ),
                            datetime.datetime.fromtimestamp(
                                float(1600443582510) / 1e6
                            ),
                            datetime.datetime.fromtimestamp(
                                float(1600443582510) / 1e6
                            ),
                        ],
                    },
                ),
                'a TIMESTAMP(6)',
                id='scalar_datetime_us',
            ),
            pytest.param(
                pd.DataFrame(
                    [
                        {'ary': [2, 3, 4]},
                        {'ary': [4444]},
                        {'ary': []},
                        {'ary': None},
                        {'ary': [2, 3, 4]},
                    ]
                ),
                'ary INT[]',
                id='array_values',
            ),
            pytest.param(
                pd.DataFrame(
                    [
                        {'ary': [2, 3, 4], 'strtest': 'teststr'},
                        {'ary': None, 'strtest': 'teststr'},
                        {'ary': [4444], 'strtest': 'teststr'},
                        {'ary': [], 'strtest': 'teststr'},
                        {'ary': [2, 3, 4], 'strtest': 'teststr'},
                    ]
                ),
                'ary INT[], strtest TEXT',
                id='mix_scalar_array_values_with_none_and_empty_list',
            ),
            pytest.param(
                gpd.GeoDataFrame(
                    {
                        'a': [Point(0, 0), Point(1, 1)],
                        'b': [
                            LineString([(2, 0), (2, 4), (3, 4)]),
                            LineString([(0, 0), (1, 1)]),
                        ],
                        'c': [
                            Polygon([(0, 0), (1, 0), (0, 1), (0, 0)]),
                            Polygon([(0, 0), (4, 0), (4, 4), (0, 4), (0, 0)]),
                        ],
                        'd': [
                            MultiPolygon(
                                [
                                    Polygon([(0, 0), (1, 0), (0, 1), (0, 0)]),
                                    Polygon(
                                        [
                                            (0, 0),
                                            (4, 0),
                                            (4, 4),
                                            (0, 4),
                                            (0, 0),
                                        ]
                                    ),
                                ]
                            ),
                            MultiPolygon(
                                [
                                    Polygon(
                                        [
                                            (0, 0),
                                            (4, 0),
                                            (4, 4),
                                            (0, 4),
                                            (0, 0),
                                        ]
                                    ),
                                    Polygon([(0, 0), (1, 0), (0, 1), (0, 0)]),
                                ]
                            ),
                        ],
                    }
                ),
                'a POINT, b LINESTRING, c POLYGON, d MULTIPOLYGON',
                id='geo_values',
            ),
        ],
    )
    def test_load_table_columnar(self, con, tmp_table, df, table_fields):
        con.execute("create table {} ({});".format(tmp_table, table_fields))
        con.load_table_columnar(tmp_table, df)
        result = _cursor2df(con.execute('select * from {}'.format(tmp_table)))
        pd.testing.assert_frame_equal(df, result)

    def test_load_infer(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        data = pd.DataFrame(
            {
                'a': np.array([0, 1], dtype=np.int32),
                'b': np.array([1.1, 2.2], dtype=np.float32),
                'c': ['a', 'b'],
            }
        )
        con.load_table("baz", data)
        con.execute("drop table if exists baz;")

    def test_load_infer_bad(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        with pytest.raises(TypeError):
            con.load_table("baz", [], method='thing')

        con.execute("drop table if exists baz;")

    def test_infer_non_pandas(self, con):

        con.execute("drop table if exists baz;")
        con.execute("create table baz (a int, b float, c text);")

        with pytest.raises(TypeError):
            con.load_table("baz", [], method='columnar')

        con.execute("drop table if exists baz;")

    def test_load_columnar_pandas_all(self, con):

        c = con.cursor()
        c.execute('drop table if exists all_types;')
        create = textwrap.dedent(
            '''\
        create table all_types (
            boolean_ BOOLEAN,
            smallint_ SMALLINT,
            int_ INT,
            bigint_ BIGINT,
            float_ FLOAT,
            double_ DOUBLE,
            varchar_ VARCHAR(40),
            text_ TEXT,
            time_ TIME,
            timestamp_ TIMESTAMP,
            date_ DATE
        );'''
        )
        # skipping decimal for now
        c.execute(create)

        data = pd.DataFrame(
            {
                "boolean_": [True, False, True, False],
                "smallint_": np.array([0, 1, 0, 1], dtype=np.int16),
                "int_": np.array([0, 1, 0, 1], dtype=np.int32),
                "bigint_": np.array([0, 1, 0, 1], dtype=np.int64),
                "float_": np.array([0, 1, 0, 1], dtype=np.float32),
                "double_": np.array([0, 1, 0, 1], dtype=np.float64),
                "varchar_": ["a", "b", "a", "b"],
                "text_": ['a', 'b', 'a', 'b'],
                "time_": [
                    datetime.time(0, 11, 59),
                    datetime.time(13),
                    datetime.time(22, 58, 59),
                    datetime.time(7, 13, 43),
                ],
                "timestamp_": [
                    pd.Timestamp("2016"),
                    pd.Timestamp("2017"),
                    pd.Timestamp(
                        '2017-11-28 23:55:59.342380', tz='US/Eastern'
                    ),
                    pd.Timestamp(
                        '2018-11-28 23:55:59.342380', tz='Asia/Calcutta'
                    ),
                ],
                "date_": [
                    datetime.date(2016, 1, 1),
                    datetime.date(2017, 1, 1),
                    datetime.date(2017, 11, 28),
                    datetime.date(2018, 11, 28),
                ],
            },
            columns=[
                'boolean_',
                'smallint_',
                'int_',
                'bigint_',
                'float_',
                'double_',
                'varchar_',
                'text_',
                'time_',
                'timestamp_',
                'date_',
            ],
        )
        con.load_table_columnar("all_types", data, preserve_index=False)

        result = list(c.execute("select * from all_types"))
        expected = [
            (
                1,
                0,
                0,
                0,
                0.0,
                0.0,
                'a',
                'a',
                datetime.time(0, 11, 59),
                datetime.datetime(2016, 1, 1, 0, 0),
                datetime.date(2016, 1, 1),
            ),
            (
                0,
                1,
                1,
                1,
                1.0,
                1.0,
                'b',
                'b',
                datetime.time(13, 0),
                datetime.datetime(2017, 1, 1, 0, 0),
                datetime.date(2017, 1, 1),
            ),
            (
                1,
                0,
                0,
                0,
                0.0,
                0.0,
                'a',
                'a',
                datetime.time(22, 58, 59),
                datetime.datetime(2017, 11, 29, 4, 55, 59),
                datetime.date(2017, 11, 28),
            ),
            (
                0,
                1,
                1,
                1,
                1.0,
                1.0,
                'b',
                'b',
                datetime.time(7, 13, 43),
                datetime.datetime(2018, 11, 28, 18, 25, 59),
                datetime.date(2018, 11, 28),
            ),
        ]

        assert result == expected
        c.execute('drop table if exists all_types;')

    def test_load_table_columnar_arrow_all(self, con):

        c = con.cursor()
        c.execute('drop table if exists all_types;')
        create = textwrap.dedent(
            '''\
        create table all_types (
            boolean_ BOOLEAN,
            smallint_ SMALLINT,
            int_ INT,
            bigint_ BIGINT,
            float_ FLOAT,
            double_ DOUBLE,
            varchar_ VARCHAR(40),
            text_ TEXT,
            time_ TIME,
            timestamp_ TIMESTAMP,
            date_ DATE
        );'''
        )
        # skipping decimal for now
        c.execute(create)

        names = [
            'boolean_',
            'smallint_',
            'int_',
            'bigint_',
            'float_',
            'double_',
            'varchar_',
            'text_',
            'time_',
            'timestamp_',
            'date_',
        ]

        columns = [
            pa.array([True, False, None], type=pa.bool_()),
            pa.array([1, 0, None]).cast(pa.int16()),
            pa.array([1, 0, None]).cast(pa.int32()),
            pa.array([1, 0, None]),
            pa.array([1.0, 1.1, None]).cast(pa.float32()),
            pa.array([1.0, 1.1, None]),
            # no fixed-width string
            pa.array(['a', 'b', None]),
            pa.array(['a', 'b', None]),
            (pa.array([1, 2, None]).cast(pa.int32()).cast(pa.time32('s'))),
            pa.array(
                [
                    datetime.datetime(2016, 1, 1, 12, 12, 12),
                    datetime.datetime(2017, 1, 1),
                    None,
                ]
            ),
            pa.array(
                [datetime.date(2016, 1, 1), datetime.date(2017, 1, 1), None]
            ),
        ]
        table = pa.Table.from_arrays(columns, names=names)
        con.load_table_arrow("all_types", table)
        c.execute('drop table if exists all_types;')

    def test_select_null(self, con):
        con.execute("drop table if exists pymapd_test_table;")
        con.execute("create table pymapd_test_table (a int);")
        con.execute("insert into pymapd_test_table VALUES (1);")
        con.execute("insert into pymapd_test_table VALUES (null);")
        # the test

        c = con.cursor()
        result = c.execute("select * from pymapd_test_table")
        expected = [(1,), (None,)]
        assert result.fetchall() == expected

        # cleanup
        con.execute("drop table if exists pymapd_test_table;")

    @pytest.mark.parametrize(
        'df, expected',
        [
            (
                pd.DataFrame(
                    {
                        "a": [1, 2],
                        "b": [1.0, 2.0],
                        "c": [
                            datetime.date(2016, 1, 1),
                            datetime.date(2017, 1, 1),
                        ],
                        "d": [
                            np.datetime64("2010-01-01T01:01:01.001001001"),
                            np.datetime64("2011-01-01T01:01:01.001001001"),
                        ],
                    }
                ),
                {
                    'a': {'type': 'BIGINT', 'is_array': False},
                    'b': {'type': 'DOUBLE', 'is_array': False},
                    'c': {'type': 'DATE', 'is_array': False},
                    'd': {
                        'type': 'TIMESTAMP',
                        'is_array': False,
                        'precision': 9,
                    },
                },
            ),
            (
                pd.DataFrame(
                    {
                        'a': [[1, 2], [1, 2], None, []],
                        'b': ['A', 'B', 'C', 'D'],
                        'c': [[1.0, 2.2], [1.0, 2.2], [], None],
                        'd': [
                            [
                                9007199254740991,
                                9007199254740992,
                                9007199254740993,
                            ],
                            [],
                            None,
                            [
                                9007199254740994,
                                9007199254740995,
                                9007199254740996,
                            ],
                        ],
                    }
                ),
                {
                    'a': {'type': 'BIGINT', 'is_array': True},
                    'b': {'type': 'STR', 'is_array': False},
                    'c': {'type': 'DOUBLE', 'is_array': True},
                    'd': {'type': 'BIGINT', 'is_array': True},
                },
            ),
            (
                gpd.GeoDataFrame(
                    {
                        'a': [Point(0, 0), Point(1, 1)],
                        'b': [
                            LineString([(2, 0), (2, 4), (3, 4)]),
                            LineString([(0, 0), (1, 1)]),
                        ],
                        'c': [
                            Polygon([(0, 0), (1, 0), (0, 1), (0, 0)]),
                            Polygon([(0, 0), (4, 0), (4, 4), (0, 4), (0, 0)]),
                        ],
                        'd': [
                            MultiPolygon(
                                [
                                    Polygon([(0, 0), (1, 0), (0, 1), (0, 0)]),
                                    Polygon(
                                        [
                                            (0, 0),
                                            (4, 0),
                                            (4, 4),
                                            (0, 4),
                                            (0, 0),
                                        ]
                                    ),
                                ]
                            ),
                            MultiPolygon(
                                [
                                    Polygon(
                                        [
                                            (0, 0),
                                            (4, 0),
                                            (4, 4),
                                            (0, 4),
                                            (0, 0),
                                        ]
                                    ),
                                    Polygon([(0, 0), (1, 0), (0, 1), (0, 0)]),
                                ]
                            ),
                        ],
                    }
                ),
                {
                    'a': {'type': 'POINT', 'is_array': True},
                    'b': {'type': 'LINESTRING', 'is_array': True},
                    'c': {'type': 'POLYGON', 'is_array': True},
                    'd': {'type': 'MULTIPOLYGON', 'is_array': True},
                },
            ),
        ],
    )
    def test_create_table(self, con, tmp_table, df, expected):
        con.create_table(tmp_table, df)
        for col in con.get_table_details(tmp_table):
            assert expected[col.name]['type'] == col.type
            if 'precision' in expected[col.name]:
                assert expected[col.name]['precision'] == col.precision

    def test_load_table_creates(self, con):

        data = pd.DataFrame(
            {
                "boolean_": [True, False],
                "smallint_cast": np.array([0, 1], dtype=np.int8),
                "smallint_": np.array([0, 1], dtype=np.int16),
                "int_": np.array([0, 1], dtype=np.int32),
                "bigint_": np.array([0, 1], dtype=np.int64),
                "float_": np.array([0, 1], dtype=np.float32),
                "double_": np.array([0, 1], dtype=np.float64),
                "varchar_": ["a", "b"],
                "text_": ['a', 'b'],
                "time_": [datetime.time(0, 11, 59), datetime.time(13)],
                "timestamp1_": [pd.Timestamp("2016"), pd.Timestamp("2017")],
                "timestamp2_": [
                    np.datetime64("2010-01-01T01:01:01.001001001"),
                    np.datetime64("2011-01-01T01:01:01.001001001"),
                ],
                "date_": [
                    datetime.date(2016, 1, 1),
                    datetime.date(2017, 1, 1),
                ],
            },
            columns=[
                'boolean_',
                'smallint_',
                'int_',
                'bigint_',
                'float_',
                'double_',
                'varchar_',
                'text_',
                'time_',
                'timestamp1_',
                'timestamp2_',
                'date_',
            ],
        )

        con.execute("drop table if exists test_load_table_creates;")
        con.load_table("test_load_table_creates", data, create=True)
        con.execute("drop table if exists test_load_table_creates;")

    def test_array_in_result_set(self, con):

        # text
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute(
            "CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 TEXT[]);"
        )

        row = [
            ("row1", "{hello,goodbye,aloha}"),
            ("row2", "{hello2,goodbye2,aloha2}"),
        ]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [
            ('row1', ['hello', 'goodbye', 'aloha']),
            ('row2', ['hello2', 'goodbye2', 'aloha2']),
        ]

        assert ans == expected

        # int
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute(
            "CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 INT[]);"
        )

        row = [("row1", "{10,20,30}"), ("row2", "{40,50,60}")]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [('row1', [10, 20, 30]), ('row2', [40, 50, 60])]

        assert ans == expected

        # timestamp
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute(
            "CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 TIMESTAMP[], col3 TIMESTAMP(9));"
        )

        row = [
            (
                "row1",
                "{2019-03-02 00:00:00,2019-03-02 00:00:00,2019-03-02 00:00:00}",  # noqa
                "2010-01-01T01:01:01.001001001",
            ),
            (
                "row2",
                "{2019-03-02 00:00:00,2019-03-02 00:00:00,2019-03-02 00:00:00}",  # noqa
                "2011-01-01T01:01:01.001001001",
            ),
        ]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [
            (
                'row1',
                [
                    datetime.datetime(2019, 3, 2, 0, 0),
                    datetime.datetime(2019, 3, 2, 0, 0),
                    datetime.datetime(2019, 3, 2, 0, 0),
                ],
                np.datetime64("2010-01-01T01:01:01.001001001"),
            ),
            (
                'row2',
                [
                    datetime.datetime(2019, 3, 2, 0, 0),
                    datetime.datetime(2019, 3, 2, 0, 0),
                    datetime.datetime(2019, 3, 2, 0, 0),
                ],
                np.datetime64("2011-01-01T01:01:01.001001001"),
            ),
        ]
        assert ans == expected

        # date
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute(
            "CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 DATE[]);"
        )

        row = [
            ("row1", "{2019-03-02,2019-03-02,2019-03-02}"),
            ("row2", "{2019-03-02,2019-03-02,2019-03-02}"),
        ]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [
            (
                'row1',
                [
                    datetime.date(2019, 3, 2),
                    datetime.date(2019, 3, 2),
                    datetime.date(2019, 3, 2),
                ],
            ),
            (
                'row2',
                [
                    datetime.date(2019, 3, 2),
                    datetime.date(2019, 3, 2),
                    datetime.date(2019, 3, 2),
                ],
            ),
        ]

        assert ans == expected

        # time
        con.execute("DROP TABLE IF EXISTS test_lists;")
        con.execute(
            "CREATE TABLE IF NOT EXISTS test_lists \
                    (col1 TEXT, col2 TIME[]);"
        )

        row = [
            ("row1", "{23:59:00,23:59:00,23:59:00}"),
            ("row2", "{23:59:00,23:59:00,23:59:00}"),
        ]

        con.load_table_rowwise("test_lists", row)
        ans = con.execute("select * from test_lists").fetchall()

        expected = [
            (
                'row1',
                [
                    datetime.time(23, 59),
                    datetime.time(23, 59),
                    datetime.time(23, 59),
                ],
            ),
            (
                'row2',
                [
                    datetime.time(23, 59),
                    datetime.time(23, 59),
                    datetime.time(23, 59),
                ],
            ),
        ]

        assert ans == expected
        con.execute("DROP TABLE IF EXISTS test_lists;")

    def test_upload_pandas_categorical_ipc(self, con):

        con.execute("DROP TABLE IF EXISTS test_categorical;")

        df = pd.DataFrame({"A": ["a", "b", "c", "a"]})
        df["B"] = df["A"].astype('category')

        # test that table created correctly when it doesn't exist on server
        con.load_table("test_categorical", df)
        ans = con.execute("select * from test_categorical").fetchall()

        assert ans == [('a', 'a'), ('b', 'b'), ('c', 'c'), ('a', 'a')]

        assert con.get_table_details("test_categorical") == [
            ColumnDetails(
                name='A',
                type='STR',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=32,
                encoding='DICT',
                is_array=False,
            ),
            ColumnDetails(
                name='B',
                type='STR',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=32,
                encoding='DICT',
                is_array=False,
            ),
        ]

        # load row-wise
        con.load_table("test_categorical", df, method="rows")

        # load columnar
        con.load_table("test_categorical", df, method="columnar")

        # load arrow
        con.load_table("test_categorical", df, method="arrow")

        # test end result
        df_ipc = con.select_ipc("select * from test_categorical")
        assert df_ipc.shape == (16, 2)

        res = df.append([df, df, df]).reset_index(drop=True)
        res["A"] = res["A"].astype('category')
        res["B"] = res["B"].astype('category')
        assert pd.DataFrame.equals(df_ipc, res)

        # test that input df wasn't mutated
        # original input is object, categorical
        # to load via Arrow, converted internally to object, object
        assert is_object_dtype(df["A"])
        assert is_categorical_dtype(df["B"])
        con.execute("DROP TABLE IF EXISTS test_categorical;")

    def test_insert_unicode(self, con):

        """INSERT Unicode using bind_params"""

        c = con.cursor()
        c.execute('drop table if exists text_holder;')
        create = 'create table text_holder (the_text text);'
        c.execute(create)
        first = {"value": "我和我的姐姐吃米饭鸡肉"}
        second = {"value": "El camina a case en bicicleta es relajante"}

        i1 = "INSERT INTO text_holder VALUES ( :value );"

        c.execute(i1, parameters=first)
        c.execute(i1, parameters=second)

        c.execute('drop table if exists text_holder;')

    def test_execute_leading_space_and_params(self, con):

        # https://github.com/omnisci/pymapd/issues/263

        """Ensure that leading/trailing spaces in execute statements
           don't cause issues
        """

        c = con.cursor()
        c.execute('drop table if exists test_leading_spaces;')
        create = 'create table test_leading_spaces (the_text text);'
        c.execute(create)
        first = {"value": "我和我的姐姐吃米饭鸡肉"}
        second = {"value": "El camina a case en bicicleta es relajante"}

        i1 = """

                    INSERT INTO test_leading_spaces


                    VALUES ( :value );

                            """

        c.execute(i1, parameters=first)
        c.execute(i1, parameters=second)

        c.execute('drop table if exists test_leading_spaces;')



import datetime
import math

import numpy as np
import pandas as pd
import pyarrow as pa
import shapely

from pandas.api.types import (
    is_bool_dtype,
    is_integer_dtype,
    is_float_dtype,
    is_object_dtype,
    is_datetime64_any_dtype,
)

from omnisci.thrift.ttypes import TColumn, TColumnData, TColumnType

from omnisci.common.ttypes import TTypeInfo, TDatumType

from ._utils import (
    date_to_seconds,
    time_to_seconds,
    datetime_to_seconds,
    mapd_to_na,
    mapd_to_slot,
)

try:
    import geopandas as gpd
except ImportError:
    gpd = None


GEO_TYPE_NAMES = ['POINT', 'LINESTRING', 'POLYGON', 'MULTIPOLYGON']
GEO_TYPE_ID = [
    v[1] for v in TDatumType._NAMES_TO_VALUES.items() if v[0] in GEO_TYPE_NAMES
]


def get_mapd_dtype(data):
    """Get the OmniSci type"""
    if is_object_dtype(data):
        return get_mapd_type_from_object(data)
    else:
        return get_mapd_type_from_known(data.dtype)


def get_mapd_type_from_known(dtype):
    """For cases where pandas type system matches"""
    if is_bool_dtype(dtype):
        return 'BOOL'
    elif is_integer_dtype(dtype):
        if dtype.itemsize <= 1:
            return 'TINYINT'
        elif dtype.itemsize == 2:
            return 'SMALLINT'
        elif dtype.itemsize == 4:
            return 'INT'
        else:
            return 'BIGINT'
    elif is_float_dtype(dtype):
        if dtype.itemsize <= 4:
            return 'FLOAT'
        else:
            return 'DOUBLE'
    elif is_datetime64_any_dtype(dtype):
        return 'TIMESTAMP'
    elif isinstance(dtype, pd.CategoricalDtype):
        return 'STR'
    else:
        raise TypeError("Unhandled type {}".format(dtype))


def get_mapd_type_from_object(data):
    """For cases where the type system mismatches"""
    try:
        val = data.dropna().iloc[0]
    except IndexError:
        raise IndexError("Not any valid values to infer the type")

    if isinstance(val, str):
        return 'STR'
    elif isinstance(val, np.datetime64):
        return 'TIMESTAMP'
    elif isinstance(val, datetime.date):
        return 'DATE'
    elif isinstance(val, datetime.time):
        return 'TIME'
    elif isinstance(val, bool):
        return 'BOOL'
    elif isinstance(val, int):
        if data.max() >= 2147483648 or data.min() <= -2147483648:
            return 'BIGINT'
        return 'INT'
    elif isinstance(val, list):
        return 'ARRAY/{}'.format(get_mapd_dtype(pd.Series(val)))
    elif isinstance(val, shapely.geometry.Point):
        return 'POINT'
    elif isinstance(val, shapely.geometry.LineString):
        return 'LINESTRING'
    elif isinstance(val, shapely.geometry.Polygon):
        return 'POLYGON'
    elif isinstance(val, shapely.geometry.MultiPolygon):
        return 'MULTIPOLYGON'
    else:
        raise TypeError("Unhandled type {}".format(data.dtype))


def thrift_cast(data, mapd_type, scale=0, precision=0):
    """Cast data type to the expected thrift types"""

    if mapd_type == 'TIMESTAMP':
        return datetime_to_seconds(data, precision)
    elif mapd_type == 'TIME':
        return pd.Series(time_to_seconds(x) for x in data)
    elif mapd_type == 'DATE':
        data = date_to_seconds(data)
        data = data.fillna(mapd_to_na[mapd_type])
        return data.astype(int)
    elif mapd_type == 'BOOL':
        # fillna before converting to int, since int cols
        # in Pandas do not support None or NaN
        data = data.fillna(mapd_to_na[mapd_type])
        return data.astype(int)
    elif mapd_type == 'DECIMAL':
        # Multiply by 10^scale
        data = data * 10 ** scale
        # fillna and convert to int
        data = data.fillna(mapd_to_na[mapd_type])
        return data.astype(int)


def build_input_columnar(
    df, preserve_index=True, chunk_size_bytes=0, col_types=[], col_names=[]
):
    if preserve_index:
        df = df.reset_index()

    dfsize = df.memory_usage().sum()
    if chunk_size_bytes > 0:
        chunks = math.ceil(dfsize / chunk_size_bytes)
    else:
        chunks = 1

    dfs = np.array_split(df, chunks)
    cols_array = []

    for df in dfs:
        input_cols = []

        colindex = 0
        for col in col_names:
            data = df.loc[:, [col]]

            mapd_type = col_types[colindex].type
            is_array = col_types[colindex].is_array
            scale = col_types[colindex].scale
            precision = col_types[colindex].precision
            has_nulls = data[col].hasnans

            if has_nulls:
                nulls = data[col].isnull().values.tolist()
            else:
                nulls = [False] * len(df)

            if is_array:
                # Expand the dataframe so each array item has
                # its own field in the dataframe.
                data = data.iloc[:, 0].apply(pd.Series)

            if mapd_type in {'TIME', 'TIMESTAMP', 'DATE', 'BOOL'}:
                # requires a cast to integer
                for c in data:
                    data.loc[:, c] = thrift_cast(
                        data=data[c], mapd_type=mapd_type, precision=precision
                    )

            if mapd_type in ['DECIMAL']:
                # requires a calculation be done using the scale
                # then cast to int
                for c in data:
                    data.loc[:, c] = thrift_cast(
                        data=data[c], mapd_type=mapd_type, scale=scale,
                    )

            if has_nulls:
                if not is_array:
                    for c in data:
                        data.loc[:, c] = data[c].fillna(mapd_to_na[mapd_type])

            if is_array:
                data = data.apply(lambda x: [i for i in x.dropna()], axis=1)
                if has_nulls:
                    data[nulls] = mapd_to_na[mapd_type]

            if mapd_type in GEO_TYPE_NAMES:
                for c in data:
                    data.loc[:, c] = data.loc[:, c].apply(lambda g: g.wkt)
            elif mapd_type not in ['FLOAT', 'DOUBLE', 'VARCHAR', 'STR']:
                if is_array:
                    data = data.apply(
                        lambda _array: [int(item) for item in _array]
                        if isinstance(_array, list)
                        else None
                    )
                else:
                    for c in data:
                        data.loc[:, c] = data.loc[:, c].astype('int64')

            # If this is an array column, we need the data to be a series
            # of TColumn objects of type mapd_type.
            if is_array:
                data = data.apply(
                    lambda x: TColumn(
                        data=TColumnData(**{mapd_to_slot[mapd_type]: x})
                    )
                )
                kwargs = {'arr_col': data}
            else:
                kwargs = {mapd_to_slot[mapd_type]: data.iloc[:, 0].values}

            input_cols.append(TColumn(data=TColumnData(**kwargs), nulls=nulls))
            colindex += 1
        cols_array.append(input_cols)

    return cols_array


def _serialize_arrow_payload(data, table_metadata, preserve_index=True):

    if isinstance(data, pd.DataFrame):

        # detect if there are categorical columns in dataframe
        cols = data.select_dtypes(include=['category']).columns

        # if there are categorical columns, make a copy before casting
        # to avoid mutating input data
        # https://github.com/omnisci/pymapd/issues/169
        if cols.size > 0:
            data_ = data.copy()
            data_[cols] = data_[cols].astype('object')
        else:
            data_ = data

        data = pa.RecordBatch.from_pandas(data_, preserve_index=preserve_index)

    stream = pa.BufferOutputStream()
    writer = pa.RecordBatchStreamWriter(stream, data.schema)

    if isinstance(data, pa.RecordBatch):
        writer.write_batch(data)
    elif isinstance(data, pa.Table):
        writer.write_table(data)

    writer.close()
    return stream.getvalue()


def build_row_desc(data, preserve_index=False):

    if not (
        isinstance(data, pd.DataFrame)
        or (gpd is not None and isinstance(data, gpd.GeoDataFrame))
    ):
        # Once https://issues.apache.org/jira/browse/ARROW-1576 is complete
        # we can support pa.Table here too
        raise TypeError(
            "Create table is not supported for type {}. "
            "Use a pandas DataFrame or a GeoPandas DataFrame, "
            "or perform the create separately".format(type(data))
        )

    if preserve_index:
        data = data.reset_index()

    dtypes = []
    is_array = {}
    for col in data.columns:
        _dtype = get_mapd_dtype(data[col])
        is_array[col] = True if _dtype.startswith('ARRAY') else None
        dtypes.append((col, _dtype.replace('ARRAY/', '')))
    # row_desc :: List<TColumnType>
    row_desc = [
        TColumnType(
            name,
            TTypeInfo(getattr(TDatumType, mapd_type), is_array=is_array[name]),
        )
        for name, mapd_type in dtypes
    ]

    # force text encoding dict for all string columns
    # default is TEXT ENCODING DICT(32) when only tct.col_type.encoding = 4 set
    # https://github.com/omnisci/pymapd/issues/140#issuecomment-477353420
    for tct in row_desc:
        if tct.col_type.type == 6:
            tct.col_type.encoding = 4
        elif tct.col_type.type in GEO_TYPE_ID:
            tct.col_type.precision = 23
        elif tct.col_type.type == 8:
            # force precision for timestamp with nanoseconds
            if data[tct.col_name].dt.nanosecond.sum():
                tct.col_type.precision = 9
    return row_desc



import datetime
import numpy as np
import pandas as pd
from enum import Enum


class TimePrecision(Enum):
    SECONDS = 0
    MILLISECONDS = 3
    MICROSECONDS = 6
    NANOSECONDS = 9


def seconds_to_time(seconds):
    """Convert seconds since midnight to a datetime.time"""
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return datetime.time(h, m, s)


def time_to_seconds(time):
    """Convert a datetime.time to seconds since midnight"""
    if time is None:
        return None
    return 3600 * time.hour + 60 * time.minute + time.second


def datetime_to_seconds(arr, precision):
    """Convert an array of datetime64[ns] to seconds since the UNIX epoch"""

    p = TimePrecision(precision)
    if arr.dtype != np.dtype('datetime64[ns]'):
        if arr.dtype == 'int64':
            # The user has passed a unix timestamp already
            return arr

        if not (
            arr.dtype == 'object'
            or str(arr.dtype).startswith('datetime64[ns,')
        ):
            raise TypeError(
                f"Invalid dtype '{arr.dtype}', expected one of: "
                "datetime64[ns], int64 (UNIX epoch), "
                "or object (string)"
            )

        # Convert to datetime64[ns] from string
        # Or from datetime with timezone information
        # Return timestamp in 'UTC'
        arr = pd.to_datetime(arr, utc=True)
        return arr.view('i8') // 10 ** 9  # ns -> s since epoch
    else:
        if p == TimePrecision.SECONDS:
            return arr.view('i8') // 10 ** 9
        elif p == TimePrecision.MILLISECONDS:
            return arr.view('i8') // 10 ** 6
        elif p == TimePrecision.MICROSECONDS:
            return arr.view('i8') // 10 ** 3
        elif p == TimePrecision.NANOSECONDS:
            return arr.view('i8')


def datetime_in_precisions(epoch, precision):
    """Convert epoch time value into s, ms, us, ns"""
    p = TimePrecision(precision)
    if p == TimePrecision.SECONDS:
        return np.datetime64(epoch, 's').astype(datetime.datetime)
    elif p == TimePrecision.MILLISECONDS:
        return np.datetime64(epoch, 'ms')
    elif p == TimePrecision.MICROSECONDS:
        return np.datetime64(epoch, 'us')
    elif p == TimePrecision.NANOSECONDS:
        return np.datetime64(epoch, 'ns')
    else:
        raise TypeError("Invalid timestamp precision: {}".format(precision))


def date_to_seconds(arr):
    """Converts date into seconds"""

    return arr.apply(lambda x: np.datetime64(x, "s").astype(int))


mapd_to_slot = {
    'BOOL': 'int_col',
    'BOOLEAN': 'int_col',
    'SMALLINT': 'int_col',
    'INT': 'int_col',
    'INTEGER': 'int_col',
    'BIGINT': 'int_col',
    'FLOAT': 'real_col',
    'DECIMAL': 'int_col',
    'DOUBLE': 'real_col',
    'TIMESTAMP': 'int_col',
    'DATE': 'int_col',
    'TIME': 'int_col',
    'STR': 'str_col',
    'POINT': 'str_col',
    'LINESTRING': 'str_col',
    'POLYGON': 'str_col',
    'MULTIPOLYGON': 'str_col',
    'TINYINT': 'int_col',
    'GEOMETRY': 'str_col',
    'GEOGRAPHY': 'str_col',
}


mapd_to_na = {
    'BOOL': -128,
    'BOOLEAN': -128,
    'SMALLINT': -32768,
    'INT': -2147483648,
    'INTEGER': -2147483648,
    'BIGINT': -9223372036854775808,
    'FLOAT': 0,
    'DECIMAL': 0,
    'DOUBLE': 0,
    'TIMESTAMP': -9223372036854775808,
    'DATE': -9223372036854775808,
    'TIME': -9223372036854775808,
    'STR': '',
    'POINT': '',
    'LINESTRING': '',
    'POLYGON': '',
    'MULTIPOLYGON': '',
    'TINYINT': -128,
    'GEOMETRY': '',
    'GEOGRAPHY': '',
}



from decimal import Decimal
import datetime

import pytest

from pymapd._loaders import _build_input_rows
from pymapd import _pandas_loaders
from pymapd._parsers import ColumnDetails
from omnisci.thrift.OmniSci import (
    TStringRow,
    TStringValue,
    TColumn,
    TColumnData,
)
import geopandas as gpd
import pandas as pd
import numpy as np
from shapely.geometry import Point, LineString, Polygon, MultiPolygon
from omnisci.thrift.ttypes import TColumnType
from omnisci.common.ttypes import TTypeInfo


def assert_columnar_equal(result, expected):

    assert len(result) == len(expected)

    for i, (a, b) in enumerate(zip(result, expected)):
        np.testing.assert_array_equal(a.nulls, b.nulls)
        np.testing.assert_array_equal(a.data.int_col, b.data.int_col)
        np.testing.assert_array_equal(a.data.real_col, b.data.real_col)
        np.testing.assert_array_equal(a.data.str_col, b.data.str_col)
        np.testing.assert_array_equal(a.data.arr_col, b.data.arr_col)


def get_col_types(col_properties: dict):
    common_col_params = dict(
        nullable=True,
        precision=0,
        scale=0,
        comp_param=0,
        encoding='NONE',
        # is_array=True,
    )

    return [
        ColumnDetails(**{**common_col_params, **properties})
        for properties in col_properties
    ]


def get_expected(data, col_properties):
    expected = []
    _map_col_types = {
        'INT': 'int_col',
        'DOUBLE': 'real_col',
        'STR': 'str_col',
        'TIMESTAMP': 'int_col',
        'DECIMAL': 'int_col',
    }
    _map_col_types.update(
        {k: 'str_col' for k in _pandas_loaders.GEO_TYPE_NAMES}
    )
    isnull = data.isnull()

    for prop in col_properties:
        nulls = isnull[prop['name']].tolist()
        if prop['is_array']:
            arr_col = []
            for v in data[prop['name']]:
                arr_col.append(
                    TColumn(
                        data=TColumnData(**{_map_col_types[prop['type']]: v})
                    ),
                )
            col = TColumn(data=TColumnData(arr_col=arr_col), nulls=nulls)
        elif prop['type'] in _pandas_loaders.GEO_TYPE_NAMES:
            col = TColumn(
                data=TColumnData(
                    **{
                        _map_col_types[prop['type']]: data[prop['name']].apply(
                            lambda g: g.wkt
                        )
                    }
                ),
                nulls=nulls,
            )
        else:
            if prop['type'] == 'TIMESTAMP':
                # convert datetime to epoch
                if data[prop['name']].dt.nanosecond.sum():
                    data[prop['name']] = data[prop['name']].astype(int)
                else:
                    data[prop['name']] = (
                        data[prop['name']].astype(int) // 10 ** 9
                    )
            elif prop['type'] == 'DECIMAL':
                # data = (data * 10 ** precision).astype(int) \
                #   * 10 ** (scale - precision)
                data[prop['name']] = (
                    data[prop['name']] * 10 ** prop['precision']
                ).astype(int) * 10 ** (prop['scale'] - prop['precision'])

            col = TColumn(
                data=TColumnData(
                    **{_map_col_types[prop['type']]: data[prop['name']]}
                ),
                nulls=nulls,
            )
        expected.append(col)
    return expected


class TestLoaders:
    def test_build_input_rows(self):
        dt_microsecond_format = '%Y-%m-%d %H:%M:%S.%f'

        def get_dt_nanosecond(v):
            return np.datetime64('201{}-01-01 01:01:01.001001001'.format(v))

        def get_dt_microsecond(v):
            return datetime.datetime.strptime(
                '201{}-01-01 01:01:01.001001'.format(v), dt_microsecond_format
            )

        data = [
            (1, 'a', get_dt_nanosecond(1), get_dt_microsecond(1)),
            (2, 'b', get_dt_nanosecond(2), get_dt_microsecond(2)),
        ]
        result = _build_input_rows(data)
        # breakpoint
        expected = [
            TStringRow(
                cols=[
                    TStringValue(str_val='1', is_null=None),
                    TStringValue(str_val='a', is_null=None),
                    TStringValue(
                        str_val=get_dt_nanosecond(1).astype(str), is_null=None
                    ),
                    TStringValue(
                        str_val=get_dt_microsecond(1).strftime(
                            dt_microsecond_format
                        ),
                        is_null=None,
                    ),
                ]
            ),
            TStringRow(
                cols=[
                    TStringValue(str_val='2', is_null=None),
                    TStringValue(str_val='b', is_null=None),
                    TStringValue(
                        str_val=get_dt_nanosecond(2).astype(str), is_null=None
                    ),
                    TStringValue(
                        str_val=get_dt_microsecond(2).strftime(
                            dt_microsecond_format
                        ),
                        is_null=None,
                    ),
                ]
            ),
        ]

        assert result == expected

    def test_build_input_rows_with_array(self):
        data = [(1, 'a'), (2, 'b'), (3, ['c', 'd', 'e'])]
        result = _build_input_rows(data)
        expected = [
            TStringRow(
                cols=[
                    TStringValue(str_val='1', is_null=None),
                    TStringValue(str_val='a', is_null=None),
                ]
            ),
            TStringRow(
                cols=[
                    TStringValue(str_val='2', is_null=None),
                    TStringValue(str_val='b', is_null=None),
                ]
            ),
            TStringRow(
                cols=[
                    TStringValue(str_val='3', is_null=None),
                    TStringValue(str_val='{c,d,e}', is_null=None),
                ]
            ),
        ]

        assert result == expected

    @pytest.mark.parametrize(
        'data, col_properties',
        [
            pytest.param(
                pd.DataFrame(
                    {
                        'a': [[1, 1], [2, 2], [3, 3]],
                        'b': [[1.1, 1.1], [2.2, 2.2], [3.3, 3.3]],
                        'c': [1, 2, 3],
                        'd': [
                            np.datetime64('2010-01-01 01:01:01.001001001'),
                            np.datetime64('2011-01-01 01:01:01.001001001'),
                            np.datetime64('2012-01-01 01:01:01.001001001'),
                        ],
                        'e': [
                            datetime.datetime.strptime(
                                '2010-01-01 01:01:01.001001',
                                '%Y-%m-%d %H:%M:%S.%f',
                            ),
                            datetime.datetime.strptime(
                                '2011-01-01 01:01:01.001001',
                                '%Y-%m-%d %H:%M:%S.%f',
                            ),
                            datetime.datetime.strptime(
                                '2012-01-01 01:01:01.001001',
                                '%Y-%m-%d %H:%M:%S.%f',
                            ),
                        ],
                        'f': [
                            Decimal('1.1234'),
                            Decimal('2.2345'),
                            Decimal('3.3456'),
                        ],
                    }
                ),
                [
                    {'name': 'a', 'type': 'INT', 'is_array': True},
                    {'name': 'b', 'type': 'DOUBLE', 'is_array': True},
                    {'name': 'c', 'type': 'INT', 'is_array': False},
                    {
                        'name': 'd',
                        'type': 'TIMESTAMP',
                        'is_array': False,
                        'precision': 9,
                    },
                    {
                        'name': 'e',
                        'type': 'TIMESTAMP',
                        'is_array': False,
                        'precision': 0,
                    },
                    {
                        'name': 'f',
                        'type': 'DECIMAL',
                        'is_array': False,
                        'scale': 10,
                        'precision': 4,
                    },
                ],
                id='mult-cols-mix-array-not-null',
            ),
            pytest.param(
                pd.DataFrame(
                    {
                        'a': [[1, 1], [2, 2], [3, 3]],
                        'b': [[1.1, 1.1], [2.2, 2.2], [3.3, 3.3]],
                    }
                ),
                [
                    {'name': 'a', 'type': 'INT', 'is_array': True},
                    {'name': 'b', 'type': 'DOUBLE', 'is_array': True},
                ],
                id='mult-cols-array-not-null',
            ),
            pytest.param(
                pd.DataFrame({'a': [1, 2, 3], 'b': [1.1, 2.2, 3.3]}),
                [
                    {'name': 'a', 'type': 'INT', 'is_array': False},
                    {'name': 'b', 'type': 'DOUBLE', 'is_array': False},
                ],
                id='mult-cols-not-null',
            ),
            pytest.param(
                pd.DataFrame(
                    [
                        {'a': [2, 3, 4]},
                        {'a': [4444]},
                        {'a': []},
                        {'a': []},
                        {'a': [2, 3, 4]},
                    ]
                ),
                [{'name': 'a', 'type': 'INT', 'is_array': True}],
                id='one-col-array-not-null',
            ),
            pytest.param(
                pd.DataFrame(
                    data=[
                        {'a': [2, 3, 4], 'b': 'teststr'},
                        {'a': [2, 3], 'b': 'teststr'},
                        {'a': [4444], 'b': 'teststr'},
                        {'a': [], 'b': 'teststr'},
                        {'a': [2, 3, 4], 'b': 'teststr'},
                    ]
                ),
                [
                    {'name': 'a', 'type': 'INT', 'is_array': True},
                    {'name': 'b', 'type': 'STR', 'is_array': False},
                ],
                id='mult-cols-mix-array-not-null',
            ),
            pytest.param(
                pd.DataFrame(
                    data=[
                        {'a': [2, 3, 4], 'b': 'teststr'},
                        {'a': [2, 3], 'b': 'teststr'},
                        {'a': [4444], 'b': 'teststr'},
                        {'a': None, 'b': 'teststr'},
                        {'a': [2, 3, 4], 'b': 'teststr'},
                    ]
                ),
                [
                    {'name': 'a', 'type': 'INT', 'is_array': True},
                    {'name': 'b', 'type': 'STR', 'is_array': False},
                ],
                id='mult-cols-mix-array-nullable',
            ),
            pytest.param(
                pd.DataFrame(
                    [
                        {'a': [2, 3, 4]},
                        {'a': [4444]},
                        {'a': None},
                        {'a': []},
                        {'a': [2, 3, 4]},
                    ]
                ),
                [{'name': 'a', 'type': 'INT', 'is_array': True}],
                id='one-col-array-nullable-and-empty-list',
            ),
            pytest.param(
                gpd.GeoDataFrame(
                    {
                        'a': [Point(0, 0), Point(1, 1)],
                        'b': [
                            LineString([(2, 0), (2, 4), (3, 4)]),
                            LineString([(0, 0), (1, 1)]),
                        ],
                        'c': [
                            Polygon([(0, 0), (1, 1), (1, 0)]),
                            Polygon([[0, 0], [0, 4], [4, 4], [4, 0]]),
                        ],
                        'd': [
                            MultiPolygon([Polygon([(0, 0), (1, 1), (1, 0)])]),
                            MultiPolygon(
                                [Polygon([[0, 0], [0, 4], [4, 4], [4, 0]])]
                            ),
                        ],
                    }
                ),
                [
                    {'name': 'a', 'type': 'POINT', 'is_array': False},
                    {'name': 'b', 'type': 'LINESTRING', 'is_array': False},
                    {'name': 'c', 'type': 'POLYGON', 'is_array': False},
                    {'name': 'd', 'type': 'MULTIPOLYGON', 'is_array': False},
                ],
                id='multi-col-geo-nullable',
            ),
        ],
    )
    def test_build_table_columnar(self, data, col_properties):

        from pymapd._pandas_loaders import build_input_columnar

        col_types = get_col_types(col_properties)

        result = build_input_columnar(
            data,
            preserve_index=False,
            col_names=data.columns,
            col_types=col_types,
        )
        expected = get_expected(data, col_properties)

        assert data.shape[1] == len(expected)
        assert_columnar_equal(result[0], expected)

    def test_build_table_columnar_pandas(self):
        common_col_params = dict(
            nullable=True,
            precision=0,
            scale=0,
            comp_param=0,
            encoding='NONE',
            is_array=False,
        )

        col_types = [
            ColumnDetails(name='boolean_', type='BOOL', **common_col_params),
            ColumnDetails(
                name='smallint_', type='SMALLINT', **common_col_params
            ),
            ColumnDetails(name='int_', type='INT', **common_col_params),
            ColumnDetails(name='bigint_', type='BIGINT', **common_col_params),
            ColumnDetails(name='float_', type='FLOAT', **common_col_params),
            ColumnDetails(name='double_', type='DOUBLE', **common_col_params),
            ColumnDetails(name='varchar_', type='STR', **common_col_params),
            ColumnDetails(name='text_', type='STR', **common_col_params),
            ColumnDetails(name='time_', type='TIME', **common_col_params),
            ColumnDetails(
                name='timestamp_',
                type='TIMESTAMP',
                nullable=True,
                precision=0,
                scale=0,
                comp_param=0,
                encoding='NONE',
                is_array=False,
            ),
            ColumnDetails(name='date_', type='DATE', **common_col_params),
        ]

        data = pd.DataFrame(
            {
                'boolean_': [True, False],
                'smallint_': np.array([0, 1], dtype=np.int16),
                'int_': np.array([0, 1], dtype=np.int32),
                'bigint_': np.array([0, 1], dtype=np.int64),
                'float_': np.array([0, 1], dtype=np.float32),
                'double_': np.array([0, 1], dtype=np.float64),
                'varchar_': ['a', 'b'],
                'text_': ['a', 'b'],
                'time_': [datetime.time(0, 11, 59), datetime.time(13)],
                'timestamp_': [pd.Timestamp('2016'), pd.Timestamp('2017')],
                'date_': [
                    datetime.date(2016, 1, 1),
                    datetime.date(2017, 1, 1),
                ],
            }
        )
        result = _pandas_loaders.build_input_columnar(
            data,
            preserve_index=False,
            col_names=data.columns,
            col_types=col_types,
        )

        nulls = [False, False]
        expected = [
            TColumn(TColumnData(int_col=[True, False]), nulls=nulls),
            TColumn(
                TColumnData(int_col=np.array([0, 1], dtype=np.int16)),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(int_col=np.array([0, 1], dtype=np.int32)),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(int_col=np.array([0, 1], dtype=np.int64)),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(real_col=np.array([0, 1], dtype=np.float32)),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(real_col=np.array([0, 1], dtype=np.float64)),
                nulls=nulls,
            ),  # noqa
            TColumn(TColumnData(str_col=['a', 'b']), nulls=nulls),
            TColumn(TColumnData(str_col=['a', 'b']), nulls=nulls),
            TColumn(TColumnData(int_col=[719, 46800]), nulls=nulls),
            TColumn(
                TColumnData(int_col=[1451606400, 1483228800]), nulls=nulls
            ),  # noqa
            TColumn(
                TColumnData(int_col=[1451606400, 1483228800]), nulls=nulls
            ),
        ]
        assert_columnar_equal(result[0], expected)

    def test_build_table_columnar_nulls(self):
        common_col_params = dict(
            nullable=True,
            scale=0,
            comp_param=0,
            encoding='NONE',
            is_array=False,
        )

        col_types = [
            ColumnDetails(
                name='boolean_', type='BOOL', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='int_', type='INT', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='bigint_', type='BIGINT', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='double_', type='DOUBLE', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='varchar_', type='STR', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='text_', type='STR', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='time_', type='TIME', precision=0, **common_col_params
            ),
            ColumnDetails(
                name='timestamp_',
                type='TIMESTAMP',
                **common_col_params,
                precision=0,
            ),
            ColumnDetails(
                name='date_', type='DATE', precision=0, **common_col_params
            ),
        ]

        data = pd.DataFrame(
            {
                'boolean_': [True, False, None],
                # Currently Pandas does not support storing None or NaN
                # in integer columns, so int cols with null
                # need to be objects. This means our type detection will be
                # unreliable since if there is no number outside the int32
                # bounds in a column with nulls then we will be assuming int
                'int_': np.array([0, 1, None], dtype=np.object),
                'bigint_': np.array(
                    [0, 9223372036854775807, None], dtype=np.object
                ),
                'double_': np.array([0, 1, None], dtype=np.float64),
                'varchar_': ['a', 'b', None],
                'text_': ['a', 'b', None],
                'time_': [datetime.time(0, 11, 59), datetime.time(13), None],
                'timestamp_': [
                    pd.Timestamp('2016'),
                    pd.Timestamp('2017'),
                    None,
                ],
                'date_': [
                    datetime.date(1001, 1, 1),
                    datetime.date(2017, 1, 1),
                    None,
                ],
            }
        )

        result = _pandas_loaders.build_input_columnar(
            data,
            preserve_index=False,
            col_names=data.columns,
            col_types=col_types,
        )

        nulls = [False, False, True]
        bool_na = -128
        int_na = -2147483648
        bigint_na = -9223372036854775808
        ns_na = -9223372037
        double_na = 0

        expected = [
            TColumn(TColumnData(int_col=[1, 0, bool_na]), nulls=nulls),
            TColumn(
                TColumnData(int_col=np.array([0, 1, int_na], dtype=np.int32)),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(
                    int_col=np.array(
                        [0, 9223372036854775807, bigint_na], dtype=np.int64
                    )
                ),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(
                    real_col=np.array([0, 1, double_na], dtype=np.float64)
                ),
                nulls=nulls,
            ),  # noqa
            TColumn(TColumnData(str_col=['a', 'b', '']), nulls=nulls),
            TColumn(TColumnData(str_col=['a', 'b', '']), nulls=nulls),
            TColumn(TColumnData(int_col=[719, 46800, bigint_na]), nulls=nulls),
            TColumn(
                TColumnData(int_col=[1451606400, 1483228800, ns_na]),
                nulls=nulls,
            ),  # noqa
            TColumn(
                TColumnData(int_col=[-30578688000, 1483228800, bigint_na]),
                nulls=nulls,
            ),  # noqa
        ]
        assert_columnar_equal(result[0], expected)

    def test_build_row_desc(self):

        data = pd.DataFrame(
            {
                'boolean_': [True, False],
                'smallint_': np.array([0, 1], dtype=np.int16),
                'int_': np.array([0, 1], dtype=np.int32),
                'bigint_': np.array([0, 1], dtype=np.int64),
                'float_': np.array([0, 1], dtype=np.float32),
                'double_': np.array([0, 1], dtype=np.float64),
                'varchar_': ['a', 'b'],
                'text_': ['a', 'b'],
                'time_': [datetime.time(0, 11, 59), datetime.time(13)],
                'timestamp1_': [pd.Timestamp('2016'), pd.Timestamp('2017')],
                'timestamp2_': [
                    np.datetime64('2016-01-01 01:01:01.001001001'),
                    np.datetime64('2017-01-01 01:01:01.001001001'),
                ],
                'date_': [
                    datetime.date(2016, 1, 1),
                    datetime.date(2017, 1, 1),
                ],
            },
            columns=[
                'boolean_',
                'smallint_',
                'int_',
                'bigint_',
                'float_',
                'double_',
                'varchar_',
                'text_',
                'time_',
                'timestamp1_',
                'timestamp2_',
                'date_',
            ],
        )
        result = _pandas_loaders.build_row_desc(data)
        expected = [
            TColumnType(
                col_name='boolean_',
                col_type=TTypeInfo(type=10),
                is_reserved_keyword=None,
            ),
            TColumnType(
                col_name='smallint_',
                col_type=TTypeInfo(type=0),
                is_reserved_keyword=None,
            ),
            TColumnType(
                col_name='int_',
                col_type=TTypeInfo(type=1),
                is_reserved_keyword=None,
            ),
            TColumnType(col_name='bigint_', col_type=TTypeInfo(type=2)),
            TColumnType(col_name='float_', col_type=TTypeInfo(type=3)),
            TColumnType(col_name='double_', col_type=TTypeInfo(type=5)),
            TColumnType(
                col_name='varchar_', col_type=TTypeInfo(type=6, encoding=4)
            ),
            TColumnType(
                col_name='text_', col_type=TTypeInfo(type=6, encoding=4)
            ),
            TColumnType(col_name='time_', col_type=TTypeInfo(type=7)),
            TColumnType(col_name='timestamp1_', col_type=TTypeInfo(type=8)),
            TColumnType(
                col_name='timestamp2_', col_type=TTypeInfo(type=8, precision=9)
            ),
            TColumnType(col_name='date_', col_type=TTypeInfo(type=9)),
        ]

        assert result == expected

        data.index.name = 'idx'
        result = _pandas_loaders.build_row_desc(data, preserve_index=True)
        expected.insert(
            0, TColumnType(col_name='idx', col_type=TTypeInfo(type=2))
        )

        assert result == expected

    def test_create_non_pandas_raises(self):
        with pytest.raises(TypeError) as m:
            _pandas_loaders.build_row_desc([(1, 'a'), (2, 'b')])

        assert m.match('is not supported for type ')



"""
The intent of this file is to be a full integration test. Whenever possible,
evaluate tests not only that a data type works, but that it works in the
presence of the other data types as well in the same dataframe/database table
"""
import pytest
import pandas as pd

from .conftest import _tests_table_no_nulls, no_gpu


@pytest.mark.usefixtures("mapd_server")
class TestGPUDataNoNulls:
    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    @pytest.mark.parametrize('method', ["rows", "columnar", "arrow", "infer"])
    def test_select_ipc_gpu(self, con, method):

        df_in = _tests_table_no_nulls(10000)
        # keep columns that pass, make issues for ones that don't
        df_in_limited = df_in[
            ["smallint_", "int_", "bigint_", "float_", "double_"]
        ].copy()

        con.execute("drop table if exists test_data_no_nulls_gpu;")
        con.load_table("test_data_no_nulls_gpu", df_in_limited, method=method)
        df_gdf = con.select_ipc_gpu("select * from test_data_no_nulls_gpu")

        # validate dtypes are exactly the same
        assert pd.DataFrame.equals(df_in_limited.dtypes, df_gdf.dtypes)

        # bring gdf local to CPU, do comparison
        df_gdf_cpu_copy = df_gdf.to_pandas()

        df_in_limited.sort_values(
            by=['smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_in_limited.reset_index(drop=True, inplace=True)

        df_gdf_cpu_copy.sort_values(
            by=['smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_gdf_cpu_copy.reset_index(drop=True, inplace=True)

        assert pd.DataFrame.equals(df_in_limited, df_gdf_cpu_copy)

        con.execute("drop table if exists test_data_no_nulls_gpu;")



import os
from codecs import open

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    'pyarrow >= 0.17,<0.18',
    'thrift == 0.13.0',
    'shapely',
    'sqlalchemy >= 1.3',
    'pandas >= 0.25,<0.26',
    'packaging >= 20.0',
    'requests >= 2.23.0',
    'numba >= 0.48',
    'rbc-project >= 0.2.2',
]

# Optional Requirements
doc_requires = ['sphinx', 'numpydoc', 'sphinx-rtd-theme']
test_requires = ['coverage', 'pytest', 'pytest-mock', 'geopandas']
dev_requires = doc_requires + test_requires + ['pre-commit']
complete_requires = dev_requires


extra_requires = {
    'docs': doc_requires,
    'test': test_requires,
    'dev': dev_requires,
    'complete': complete_requires,
}

setup(
    name='pymapd',
    description='A DB API 2 compatible client for OmniSci (formerly MapD).',
    long_description=long_description,
    url='https://github.com/omnisci/pymapd',
    author='OmniSci',
    author_email='community@omnisci.com',
    license='Apache Software License',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Database',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
    ],
    packages=find_packages(),
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    install_requires=install_requires,
    extras_require=extra_requires,
)



"""
The intent of this file is to be a full integration test. Whenever possible,
add a datatype to the main _tests_table_no_nulls function, so that the tests
will evaluate not only that a data type works, but that it works in the
presence of the other data types as well in the same dataframe/database table
"""
import pytest
import pandas as pd
import numpy as np
from shapely import wkt

from .conftest import _tests_table_no_nulls


@pytest.mark.usefixtures("mapd_server")
class TestCPUDataNoNulls:
    @pytest.mark.parametrize('method', ["rows", "columnar", "arrow", "infer"])
    def test_create_load_table_no_nulls_sql_execute(self, con, method):
        """
        Demonstrate that regardless of how data loaded, roundtrip answers
        are the same when con.execute()/pd.read_sql called for row-wise
        data retrieval
        """
        df_in = _tests_table_no_nulls(10000)
        df_in.drop(
            columns=["point_", "line_", "mpoly_", "poly_"], inplace=True
        )
        con.execute("drop table if exists test_data_no_nulls;")
        con.load_table("test_data_no_nulls", df_in, method=method)

        # read_sql() uses execute() under the hood
        df_out = pd.read_sql("select * from test_data_no_nulls", con)

        # test size and table definition
        assert df_in.shape == df_out.shape
        gtd = con.get_table_details("test_data_no_nulls")
        name_types = [(x.name, x.type) for x in gtd]
        assert name_types == [
            ('tinyint_', 'TINYINT'),
            ('smallint_', 'SMALLINT'),
            ('int_', 'INT'),
            ('bigint_', 'BIGINT'),
            ('float_', 'FLOAT'),
            ('double_', 'DOUBLE'),
            ('bool_', 'BOOL'),
            ('date_', 'DATE'),
            ('datetime_', 'TIMESTAMP'),
            ('time_', 'TIME'),
            ('text_', 'STR'),
        ]

        # sort tables to ensure data in same order before compare
        # need to sort by all the columns in case of ties
        df_in.sort_values(
            by=['tinyint_', 'smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_in.reset_index(drop=True, inplace=True)

        df_out.sort_values(
            by=['tinyint_', 'smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_out.reset_index(drop=True, inplace=True)

        # pymapd won't necessarily return exact dtype as input using execute()
        # and pd.read_sql() since transport is rows of tuples
        # test that results are the same when dtypes aligned
        assert pd.DataFrame.equals(
            df_in["tinyint_"], df_out["tinyint_"].astype('int8')
        )

        assert pd.DataFrame.equals(
            df_in["smallint_"], df_out["smallint_"].astype('int16')
        )

        assert pd.DataFrame.equals(
            df_in["int_"], df_out["int_"].astype('int32')
        )

        assert pd.DataFrame.equals(df_in["bigint_"], df_out["bigint_"])
        assert all(np.isclose(df_in["float_"], df_out["float_"]))
        assert all(np.isclose(df_in["double_"], df_out["double_"]))
        assert pd.DataFrame.equals(
            df_in["bool_"], df_out["bool_"].astype('bool')
        )  # noqa
        assert pd.DataFrame.equals(df_in["date_"], df_out["date_"])
        assert pd.DataFrame.equals(df_in["datetime_"], df_out["datetime_"])
        assert pd.DataFrame.equals(df_in["time_"], df_out["time_"])
        assert pd.DataFrame.equals(df_in["text_"], df_out["text_"])

        con.execute("drop table if exists test_data_no_nulls;")

    @pytest.mark.parametrize('method', ["rows", "columnar", "arrow", "infer"])
    def test_create_load_table_no_nulls_select_ipc(self, con, method):
        """
        Demonstrate that regardless of how data loaded, roundtrip answer
        is same when con.select_ipc() called to retrieve data using Arrow
        """
        # need to manually specify columns since some don't currently work
        # need to drop unsupported columns from df_in
        df_in = _tests_table_no_nulls(10000)
        df_in.drop(
            columns=["point_", "line_", "mpoly_", "poly_"], inplace=True
        )

        con.execute("drop table if exists test_data_no_nulls_ipc;")
        con.load_table("test_data_no_nulls_ipc", df_in, method=method)

        df_out = con.select_ipc(
            """select
                                tinyint_,
                                smallint_,
                                int_,
                                bigint_,
                                float_,
                                double_,
                                bool_,
                                date_,
                                datetime_,
                                time_,
                                text_
                                from test_data_no_nulls_ipc"""
        )

        # test size and table definition
        assert df_in.shape == df_out.shape

        # sort tables to ensure data in same order before compare
        # need to sort by all the columns in case of ties
        df_in.sort_values(
            by=['tinyint_', 'smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_in.reset_index(drop=True, inplace=True)

        df_out.sort_values(
            by=['tinyint_', 'smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_out.reset_index(drop=True, inplace=True)

        # When Arrow result converted to pandas, dict comes back as category
        # This providies extra functionality above base 'object' type
        df_out["text_"] = df_out["text_"].astype('object')

        # select_ipc uses Arrow, so expect exact df dtypes back
        assert pd.DataFrame.equals(df_in, df_out)

        con.execute("drop table if exists test_data_no_nulls_ipc;")

    @pytest.mark.parametrize('method', ["rows", "columnar"])
    def test_load_table_text_no_encoding_no_nulls(self, con, method):
        """
        Demonstrate that data can be loaded as text encoding none,
        assuming that user creates the table beforehand/inserting to
        pre-existing table
        """

        con.execute("drop table if exists test_text_no_encoding")

        con.execute(
            """create table test_text_no_encoding (
                       idx integer,
                       text_ text encoding none
                       )"""
        )

        # reset_index adds a column to sort by, since results not guaranteed
        # to return in sorted order from OmniSci
        df_in = _tests_table_no_nulls(10000)
        df_test = df_in["text_"].reset_index()
        df_test.columns = ["idx", "text_"]

        con.load_table("test_text_no_encoding", df_test, method=method)

        df_out = pd.read_sql(
            """select
                                *
                                from test_text_no_encoding
                                order by idx""",
            con,
        )

        assert pd.DataFrame.equals(df_test, df_out)

        con.execute("drop table if exists test_text_no_encoding")

    @pytest.mark.parametrize('method', ["rows"])
    def test_load_table_geospatial_no_nulls(self, con, method):
        """
        Demonstrate that geospatial data can be loaded,
        assuming that user creates the table beforehand/inserting to
        pre-existing table
        """
        con.execute("drop table if exists test_geospatial_no_nulls")

        con.execute(
            """create table test_geospatial_no_nulls (
                       tinyint_ tinyint,
                       smallint_ smallint,
                       int_ integer,
                       bigint_ bigint,
                       float_ float,
                       double_ double,
                       bool_ boolean,
                       date_ date,
                       datetime_ timestamp,
                       time_ time,
                       text_ text encoding dict(32),
                       point_ point,
                       line_ linestring,
                       mpoly_ multipolygon,
                       poly_ polygon
                       )"""
        )

        df_in = _tests_table_no_nulls(10000)
        con.load_table("test_geospatial_no_nulls", df_in, method='rows')

        df_out = pd.read_sql(
            """select
                             *
                             from test_geospatial_no_nulls""",
            con,
        )

        # sort tables to ensure data in same order before compare
        # need to sort by all the columns in case of ties
        df_in.sort_values(
            by=['tinyint_', 'smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_in.reset_index(drop=True, inplace=True)

        df_out.sort_values(
            by=['tinyint_', 'smallint_', 'int_', 'bigint_'], inplace=True
        )
        df_out.reset_index(drop=True, inplace=True)

        # pymapd won't necessarily return exact dtype as input using execute()
        # and pd.read_sql() since transport is rows of tuples
        # test that results are the same when dtypes aligned
        assert pd.DataFrame.equals(
            df_in["tinyint_"], df_out["tinyint_"].astype('int8')
        )

        assert pd.DataFrame.equals(
            df_in["smallint_"], df_out["smallint_"].astype('int16')
        )

        assert pd.DataFrame.equals(
            df_in["int_"], df_out["int_"].astype('int32')
        )

        assert pd.DataFrame.equals(df_in["bigint_"], df_out["bigint_"])
        assert all(np.isclose(df_in["float_"], df_out["float_"]))
        assert all(np.isclose(df_in["double_"], df_out["double_"]))
        assert pd.DataFrame.equals(
            df_in["bool_"], df_out["bool_"].astype('bool')
        )  # noqa
        assert pd.DataFrame.equals(df_in["date_"], df_out["date_"])
        assert pd.DataFrame.equals(df_in["datetime_"], df_out["datetime_"])
        assert pd.DataFrame.equals(df_in["time_"], df_out["time_"])
        assert pd.DataFrame.equals(df_in["text_"], df_out["text_"])

        # convert geospatial data to Shapely objects to prove their equality
        point_in = [wkt.loads(x) for x in df_in["point_"]]
        point_out = [wkt.loads(x) for x in df_out["point_"]]
        assert all(
            [x.equals_exact(y, 0.000001) for x, y in zip(point_in, point_out)]
        )

        line_in = [wkt.loads(x) for x in df_in["line_"]]
        line_out = [wkt.loads(x) for x in df_out["line_"]]
        assert all(
            [x.equals_exact(y, 0.000001) for x, y in zip(line_in, line_out)]
        )

        mpoly_in = [wkt.loads(x) for x in df_in["mpoly_"]]
        mpoly_out = [wkt.loads(x) for x in df_out["mpoly_"]]
        assert all(
            [x.equals_exact(y, 0.000001) for x, y in zip(mpoly_in, mpoly_out)]
        )

        # TODO: tol only passes at 0.011, whereas others pass at much tighter
        # Figure out why
        poly_in = [wkt.loads(x) for x in df_in["poly_"]]
        poly_out = [wkt.loads(x) for x in df_out["poly_"]]
        assert all(
            [x.equals_exact(y, 0.011) for x, y in zip(poly_in, poly_out)]
        )

        con.execute("drop table if exists test_geospatial_no_nulls")



import os
import subprocess
import pytest
from pymapd import connect
from omnisci.thrift.ttypes import TOmniSciException, TApplicationException
from .conftest import no_gpu
import pandas as pd


@pytest.mark.usefixtures("mapd_server")
class TestDeallocate:
    def _data(self):
        HERE = os.path.dirname(__file__)
        df = pd.read_pickle(os.path.join(HERE, "data", "data_load.pkl"))
        return df.itertuples(index=False)

    def _connect(self):

        return connect(
            user="admin",
            password='HyperInteractive',
            host='localhost',
            port=6274,
            protocol='binary',
            dbname='omnisci',
        )

    def _transact(self, con):
        drop = 'drop table if exists iris;'
        con.execute(drop)
        create = '''create table iris (
        id int, sepal_l float, sepal_w float,
        petal_l int, petal_w float, species text);'''
        con.execute(create)
        con.load_table('iris', self._data())

    @pytest.mark.skip(reason="deallocate non-functional in recent distros")
    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_deallocate_ipc_gpu(self):
        con = self._connect()
        self._transact(con)
        bmem = int(
            subprocess.check_output(
                '''nvidia-smi -i 0 \
                                          -q -d MEMORY | \
                                          awk ' FNR == 11 { print $3 }' \
                                          ''',
                shell=True,
            )
        )
        df = con.select_ipc_gpu(
            '''select sepal_l, sepal_W,
                                petal_l, petal_w, species from iris'''
        )
        con.deallocate_ipc_gpu(df)
        amem = int(
            subprocess.check_output(
                '''nvidia-smi -i 0 \
                                          -q -d MEMORY | \
                                          awk ' FNR == 11 { print $3 }' \
                                          ''',
                shell=True,
            )
        )
        con.close()

        assert amem == bmem

    @pytest.mark.skip(reason="deallocate non-functional in recent distros")
    def test_deallocate_ipc(self):
        con = self._connect()
        self._transact(con)
        bmem = int(subprocess.check_output('''ipcs -m | wc -l''', shell=True))
        df = con.select_ipc(
            '''select sepal_l, sepal_W,
                                petal_l, petal_w, species from iris'''
        )
        con.deallocate_ipc(df)
        amem = int(subprocess.check_output('''ipcs -m | wc -l''', shell=True))
        con.close()

        assert amem == bmem

    @pytest.mark.skip(reason="deallocate non-functional in recent distros")
    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_gpu_raises_right_Exception(self):
        con = self._connect()
        self._transact(con)

        df = con.select_ipc_gpu('select species from iris')
        con.deallocate_ipc_gpu(df)

        with pytest.raises(TOmniSciException) as te:
            con.deallocate_ipc_gpu(df)
        assert 'Exception: current data frame' in str(te.value.error_msg)

    @pytest.mark.skip(reason="deallocate non-functional in recent distros")
    def test_ipc_raises_right_exception(self):
        con = self._connect()
        self._transact(con)

        df = con.select_ipc('select sepal_l, sepal_W from iris')
        con.deallocate_ipc(df)

        with pytest.raises(TApplicationException) as ae:
            con.deallocate_ipc(df)
        assert 'valid shm ID w/ given shm key' in str(ae.value.message)

    @pytest.mark.skip(reason="deallocate non-functional in recent distros")
    @pytest.mark.skipif(no_gpu(), reason="No GPU available")
    def test_deallocate_session(self):
        con = self._connect()
        con1 = self._connect()
        self._transact(con)

        df = con.select_ipc_gpu('select id from iris')
        con.close()
        with pytest.raises(TOmniSciException) as te:
            con1.deallocate_ipc_gpu(df)
        assert 'Exception: current data frame' in str(te.value.error_msg)

    @pytest.mark.skip(reason="deallocate non-functional in recent distros")
    def test_ipc_multiple_df(self):
        con = self._connect()
        self._transact(con)
        bmem = int(subprocess.check_output('''ipcs -m | wc -l''', shell=True))
        df1 = con.select_ipc(
            '''select sepal_l, sepal_W,
                                petal_l, petal_w, species from iris'''
        )
        df2 = con.select_ipc(
            '''select sepal_l, sepal_W,
                                petal_l, petal_w, species from iris'''
        )
        df3 = con.select_ipc(
            '''select sepal_l, sepal_W,
                                petal_l, petal_w, species from iris'''
        )
        con.deallocate_ipc(df1)
        con.deallocate_ipc(df2)
        con.deallocate_ipc(df3)

        amem = int(subprocess.check_output('''ipcs -m | wc -l''', shell=True))
        con.close()

        assert amem == bmem
